/*=====================================================================
processlist.cpp
---------------
File created by ClassTemplate on Sun Mar 20 17:33:43 2005

Copyright (C) Nicholas Chapman

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

http://www.gnu.org/copyleft/gpl.html.
=====================================================================*/
#include "processlist.h"
#include <algorithm>

BEGIN_EVENT_TABLE(ProcessList, wxListCtrl)
EVT_LIST_ITEM_SELECTED(PROCESS_LIST, ProcessList::OnSelected)
EVT_LIST_COL_CLICK(-1, ProcessList::OnSort)
EVT_TIMER(PROCESS_LIST_TIMER, ProcessList::OnTimer)
END_EVENT_TABLE()

ProcessList::ProcessList(wxWindow *parent, const wxPoint& pos,
						 const wxSize& size)
						 :	wxSortedListCtrl(parent, PROCESS_LIST, pos, size, wxLC_REPORT),
						 timer(this, PROCESS_LIST_TIMER)
{
	InitSort();

	selected_process = 0;

	wxListItem itemCol;
	itemCol.m_mask = wxLIST_MASK_TEXT/* | wxLIST_MASK_IMAGE*/;
	itemCol.m_text = _T("Process");
	itemCol.m_image = -1;
	InsertColumn(COL_NAME, itemCol);
	itemCol.m_text = _T("CPU Usage");
	InsertColumn(COL_CPUUSAGE, itemCol);

	SetColumnWidth(COL_NAME, 200);
	SetColumnWidth(COL_CPUUSAGE, 100);

	sort_column = COL_CPUUSAGE;
	sort_dir = SORT_DOWN;
	SetSortImage(sort_column, sort_dir); 

	updateProcesses();
	timer.Start(1000); // 1 second interval
}


ProcessList::~ProcessList()
{

}

void ProcessList::OnSelected(wxListEvent& event)
{
	this->selected_process = event.m_itemIndex;
}

const ProcessInfo* ProcessList::getSelectedProcess()
{
	if(selected_process >= 0 && selected_process < (int)processes.size())
	{
		return &processes[selected_process];
	}
	else
		return NULL;
}

__int64 getDiff(FILETIME before, FILETIME after)
{
	__int64 i0 = (__int64(before.dwHighDateTime) << 32) + before.dwLowDateTime;
	__int64 i1 = (__int64( after.dwHighDateTime) << 32) +  after.dwLowDateTime;
	return i1 - i0;
}

void ProcessList::OnTimer(wxTimerEvent& event)
{
	updateTimes();
}

struct NameAscPred { bool operator () (const ProcessInfo &a, const ProcessInfo &b) {
	return stricmp(a.getName().c_str(), b.getName().c_str()) < 0;
} };

struct NameDescPred { bool operator () (const ProcessInfo &a, const ProcessInfo &b) {
	return stricmp(a.getName().c_str(), b.getName().c_str()) > 0;
} };

struct CpuUsageAscPred { bool operator () (const ProcessInfo &a, const ProcessInfo &b) {
	return a.cpuUsage < b.cpuUsage;
} };

struct CpuUsageDescPred { bool operator () (const ProcessInfo &a, const ProcessInfo &b) {
	return a.cpuUsage > b.cpuUsage;
} };

void ProcessList::sortByName()
{
	if (sort_dir == SORT_UP)
		std::sort(processes.begin(), processes.end(), NameAscPred());
	else
		std::sort(processes.begin(), processes.end(), NameDescPred());
}

void ProcessList::sortByCpuUsage()
{
	if (sort_dir == SORT_UP)
		std::sort(processes.begin(), processes.end(), CpuUsageAscPred());
	else
		std::sort(processes.begin(), processes.end(), CpuUsageDescPred());
}

void ProcessList::OnSort(wxListEvent& event)
{
	SetSortImage(sort_column, SORT_NONE);

	if (sort_column == event.m_col)
	{
		// toggle if we clicked on the same column as last time
		sort_dir = (SortType)((SORT_UP+SORT_DOWN) - sort_dir);
	} else {
		// if switching columns, start with the default sort for that column type
		sort_column = event.m_col;
		sort_dir = (sort_column >= 1 && sort_column <= 4) ? SORT_DOWN : SORT_UP;
	}

	SetSortImage(sort_column, sort_dir);
	updateSorting();
}

void ProcessList::updateSorting()
{
	switch(sort_column) {
		case 0: sortByName(); break;
		case 1: sortByCpuUsage(); break;
	}
	fillList();
}

void ProcessList::fillList()
{
	for(int i=0; i<(int)processes.size(); ++i)
	{
		this->SetItem(i, COL_NAME, processes[i].getName().c_str());

		char str[32];
		if (processes[i].cpuUsage >= 0)
			sprintf(str, "%i%%", processes[i].cpuUsage);
		else
			strcpy(str, "-");
		this->SetItem(i, COL_CPUUSAGE, str);
	}
}

void ProcessList::updateProcesses()
{
	DeleteAllItems();
	processes.clear();

	ProcessInfo::enumProcesses(this->processes);
	for(int i=0; i<(int)processes.size(); ++i)
	{
		long tmp = this->InsertItem(i, "", -1);
		SetItemData(tmp, i);
	}

	lastTime = wxGetLocalTimeMillis();
	updateTimes();

	// We need to wait a bit before we can get any useful CPU usage data.
	{
		int steps = 50;
		wxProgressDialog dlg("Sleepy", "Searching for processes...", steps);
		for (int n=0;n<steps;n++)
		{
			Sleep(500/steps);
			dlg.Update(n);
		}
	}

	// Now we've waited, we can grab the data again.
	updateTimes();
	updateSorting();
	fillList();
}

void ProcessList::updateTimes()
{
	wxLongLong now = wxGetLocalTimeMillis();
	int sampleTimeDiff = (now - lastTime).ToLong();
	lastTime = now;

	for(int i=0; i<(int)processes.size(); ++i)
	{
		processes[i].cpuUsage = -1;

		HANDLE process_handle = OpenProcess(PROCESS_ALL_ACCESS, FALSE, processes[i].getID()); 
		DWORD err = GetLastError();
		BOOL result = FALSE;
		FILETIME CreationTime, ExitTime, KernelTime, UserTime;
		if (process_handle != NULL)
		{
			result = GetProcessTimes(
				process_handle,
				&CreationTime,
				&ExitTime,
				&KernelTime,
				&UserTime
				);

			CloseHandle(process_handle);
		}

		if (result) 
		{
			__int64 kernel_diff = getDiff(processes[i].prevKernelTime, KernelTime);
			__int64 user_diff = getDiff(processes[i].prevUserTime, UserTime);
			processes[i].prevKernelTime = KernelTime;
			processes[i].prevUserTime = UserTime;

			if (sampleTimeDiff > 0)
				processes[i].cpuUsage = ((kernel_diff + user_diff) / 10000) * 100 / sampleTimeDiff;

		}
	}
	fillList();
}


